﻿using System;
using System.Collections.Generic;
using UnityEngine;

public class WaveManager : MonoBehaviour
{
    private const int length = 70;
    private float[,] heights;
    private Vector2[,] directions;
    private float deltaTime;
    private float updateDelta = 0.5f;
    private float lifeTime;
    private float endTime = -1;

    private Dictionary<int, Emitter> centers = new Dictionary<int, Emitter>();
    private Mesh waveMesh;
    private Vector3[] verticeList;
    public Material SurfaceMaterial;
    private System.Random random = new System.Random();

    [SerializeField] private float desiredWidth = 200;
    private bool restart;

    public float DesiredWidth
    {
        get
        {
            return desiredWidth;
        }       
    }

    public bool IsEnded { get { return endTime >= 0; } }

    public void Awake()
    {
        heights = new float[length, length];
        directions = new Vector2[length, length];
        for (int y = 0; y < length; y++)
        {
            for (int x = 0; x < length; x++)
            {
                heights[x, y] = 0f;
                directions[x, y] = Vector2.zero;
            }
        }

        BuildMesh();
        SetMesh();
    }

    public void End()
    {
        if (endTime >= 0)
            return;

        endTime = lifeTime;
        restart = false;
    }

    public void Restart()
    {
        if (restart)
            return;

        endTime = lifeTime;
        restart = true;
    }

    internal Vector2 GetForceAt(Vector2 position, out float height)
    {
        Vector2 pos = position / DesiredWidth * length + Vector2.one * length / 2;
        
        int xOffsetInt = (int)pos.x;
        int yOffsetInt = (int)pos.y;

        ClampOnCircle(xOffsetInt, yOffsetInt, out xOffsetInt, out yOffsetInt);

        var dir = directions[xOffsetInt, yOffsetInt];
        height = heights[xOffsetInt, yOffsetInt];
        return dir * height;
    }

    private void BuildMesh()
    {
        var scale = DesiredWidth / length;

        waveMesh = new Mesh();
        verticeList = new Vector3[length * length];
        List<Vector2> uvList = new List<Vector2>();
        List<int> triList = new List<int>();
        for (int y = 0; y < length; y++)
        {
            for (int x = 0; x < length; x++)
            {
                int xMapped, yMapped;
                ClampOnCircle(x, y, out xMapped, out yMapped);
                verticeList[x + y * length] = new Vector3(xMapped*scale - DesiredWidth/2, 0f, yMapped*scale - DesiredWidth / 2);
                uvList.Add(new Vector2(x, y));
                //Skip if a new square on the plane hasn't been formed
                if (x == 0 || y == 0)
                    continue;
                //Adds the index of the three vertices in order to make up each of the two tris
                triList.Add(length * (x - 1) + y - 1); //Bottom left - First triangle
                triList.Add(length * x + y - 1); //Bottom right
                triList.Add(length * x + y); //Top right
                triList.Add(length * x + y); //Top right - Second triangle
                triList.Add(length * (x - 1) + y); //Top left
                triList.Add(length * (x - 1) + y - 1); //Bottom left                
            }
        }
        waveMesh.vertices = verticeList;
        waveMesh.uv = uvList.ToArray();
        waveMesh.triangles = triList.ToArray();

        var filter = this.gameObject.GetComponent<MeshFilter>();
        filter.sharedMesh = waveMesh;

        var renderer = this.gameObject.GetComponent<MeshRenderer>();
        renderer.sharedMaterial = SurfaceMaterial;
    }

    public void Update()
    {
        var delta = Time.deltaTime;
        deltaTime += delta;
        lifeTime += delta;

        if (deltaTime < updateDelta)
            return;

        deltaTime -= delta;

        // Kill all
        if (endTime >= 0)
            foreach (var item in centers)
                if (!item.Value.IsDying(lifeTime))
                    item.Value.Decease(lifeTime);

        List<int> deadIds = new List<int>();
        foreach (var item in centers)
        {
            if (item.Value.IsDead(lifeTime))
                deadIds.Add(item.Key);
        }

        foreach (var item in deadIds)
        {
            centers.Remove(item);
        }

        var progress = Mathf.Clamp01((lifeTime - endTime) / 2f);
        if (restart)
            progress = 1 - progress;

        float suddenDeath = endTime >= 0 ? progress * 3 : 0;

        if (endTime >=0 && suddenDeath >= 1f && restart)
        {
            endTime = -1;
            restart = false;
            return;
        }

        for (int y = 0; y < length; y++)
        {
            for (int x = 0; x < length; x++)
            {
                Vector2 direction;
                heights[x, y] = ComputeHeightAt(x, y, lifeTime, out direction) + suddenDeath;
                directions[x, y] = direction;
            }
        }

        SetMesh();
    }

    private void SetMesh()
    {
        // Draw data grid
        for (int y = 0; y < heights.GetLength(0); y++)
            for (int x = 0; x < heights.GetLength(1); x++)
            {
                float power = heights[x, y] * 5;
                Vector3 value = verticeList[y * length + x];

                verticeList[y * length + x] = new Vector3(value.x, power, value.z);
            }

        this.waveMesh.vertices = verticeList;
        this.waveMesh.RecalculateNormals();
    }

    public void KillById(int id)
    {
        Emitter emitter;
        if (this.centers.TryGetValue(id, out emitter))
        {
            emitter.Decease(this.lifeTime);
        }
    }

    public void Reset()
    {
        centers.Clear();
    }

    /// <summary>
    /// Input between 0..length. ans output as well.
    /// </summary>
    private void ClampOnCircle(int xset, int yset, out int newX, out int newY)
    {
        Vector2 clampedPos = Vector2.ClampMagnitude(new Vector2(xset - length / 2, yset - length / 2), length / 2);

        newX = (int)Mathf.Clamp(clampedPos.x + length / 2, 0, length - 1);
        newY = (int)Mathf.Clamp(clampedPos.y + length / 2, 0, length - 1);
    }

    private float ComputeHeightAt(int x, int y, float lifeTime, out Vector2 direction)
    {
        float sum = 0;
        direction = Vector2.zero;
        foreach (var item in centers)
        {
            var pos = new Vector2(x, y);
            var power = item.Value.ComputeHeightAt(pos, lifeTime);
            sum += power;
            direction += item.Value.ComputeDirectionAt(pos, lifeTime) * Mathf.Max(0,power);
        }

        return sum;
    }

    public int Add(Vector2 center, float killTime = -1, float radius = 1)
    {
        int id;

        do { id = random.Next(); } while (this.centers.ContainsKey(id));

        Vector2 pos = center / DesiredWidth * length + Vector2.one * length / 2;
        this.centers.Add(id, new Emitter(pos, lifeTime, killTime < 0 ? -1 : (lifeTime + killTime), radius));

        return id;
    }


    // Gizmos from here on 

    public void OnDrawGizmosSelected()
    {
        var scale = DesiredWidth / length;


        //DrawGizmoCircle(new Vector3(transform.position.x + length / 2, transform.position.y, transform.position.z + length / 2), length / 2);

        if (heights == null)
            return;

        // Draw data grid
        for (int y = 0; y < heights.GetLength(0); y++)
            for (int x = 0; x < heights.GetLength(1); x++)
            {
                Vector2 direction = directions[x, y]*4;

                if (direction == Vector2.zero)
                    continue;

                //var value = verticeList[y * length + x];
                var from = new Vector3((x + transform.localPosition.x) * scale -  DesiredWidth/2, 5, (y + transform.localPosition.z) * scale - DesiredWidth / 2);
                var dir = new Vector3(direction.x, 0, direction.y);
                Gizmos.color = Color.red;
                Gizmos.DrawRay(from, dir);
                Gizmos.color = Color.green;
                Gizmos.DrawCube(from + dir, Vector3.one * 0.2f);

                //Gizmos.DrawCube(new Vector3(i, 0, j), new Vector3(0.1f, 0.1f, 0.1f));
                //Gizmos.DrawCube(new Vector3(i, 0, j), new Vector3(power, 1, power));
            }
    }


    private class Emitter
    {
        private Vector2 center;
        private float spawnTime;
        private float unspawnTime = -1;
        private float radius;

        public bool IsDying(float timeNow)
        {
            return unspawnTime >= 0;
        }

        public bool IsDead(float timeNow)
        {
            return unspawnTime >= 0 && ((timeNow - unspawnTime) / 0.5f) > 1;
        }

        public Emitter(Vector2 center, float lifeTime, float killTime, float radius)
        {
            this.center = center;
            this.spawnTime = lifeTime;
            this.unspawnTime = killTime;
            this.radius = radius;
        }

        public void Decease(float lifeTime)
        {
            this.unspawnTime = lifeTime;
        }

        public Vector2 ComputeDirectionAt(Vector2 checkPoint, float timeNow)
        {
            return (-center+checkPoint).normalized;
        }

        public float ComputeHeightAt(Vector2 checkPoint, float timeNow)
        {
            float result;

            float maxDistance = 20*radius;
            float maxDistanceSqr = maxDistance * maxDistance;
            float currentDistanceSqr = (checkPoint - center).sqrMagnitude;

            float spawnProgress = Mathf.Clamp01((timeNow - spawnTime) / 0.5f);

            float distanceProgress = 1 - Mathf.Clamp01(currentDistanceSqr / maxDistanceSqr);
            float timeProgress = Mathf.Clamp01((timeNow - spawnTime) / maxDistance) * 10;

            result = Mathf.Sin((distanceProgress + timeProgress) * Mathf.PI * 1.5f) * distanceProgress * spawnProgress;

            if (unspawnTime >= 0)
            {
                float lifeTimeEndProgress = Mathf.Clamp01((timeNow - unspawnTime) / 0.5f);

                result *= (1 - lifeTimeEndProgress);
            }

            return result;
        }
    }

    public void OnDestroy()
    {
        Destroy(this.waveMesh);
    }
}